/* isl2java.c */
/* Chris Jacobi, April 28, 1997 11:47 am PDT */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */

 
/* $Id: isl2java.c,v 1.40 1997/04/28 18:50:23 jacobi Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "iluptype.h"
#include "shared.h"
#include "stubops.h"
#include "name.h"
#include "util.h"
#include "genstub.h"
#include "prttree.h"
#include "context.h"


#ifdef _IS_POSIX
#include <unistd.h>
#include <errno.h>
#endif				/* _IS_POSIX */


static void
generateLoadClass(Interface ifc) {
    IHandle ih = getContext(ifc);
    char* classShortName = ih->loaderClass_Name;
    if (! classShortName) return;
    NewJavaFile(ifc, classShortName);
    printf("import xerox.basics.Environment;\n\n");
    printf("/**\n");
    printf(" * Loads all java classes used for the stub (client) \n");
    printf(" * side of the ilu interface which might not be loaded  \n");
    printf(" * automatically when needed.\n");
    printf(" */\n");
    printf("public class %s {\n",
        classShortName
        );
    printf("    public static void load(){}\n");
    printf("    static {\n");
    LOOP_BEGIN(ih->stuffToLoad, char*, className, t1)
        printf("        Environment.loadClasses(\"%s\");\n",
            dotCat(packageName(ifc), className) 
            /*no short name; this is a string */
            );
    LOOP_END()
    printf("    }\n");
    printf("}//%s\n\n", classShortName);
}


static boolean
matchImportedWithString(void *p1, void *p2)
{
    return strcmp(((Imported) p1)->name, (char *) p2) == 0;
}


static void
addInterfaceToImports(list imports, Interface ifc)
{
    char *name = ifc->name->base_name;

    if (list_find(imports, matchImportedWithString, name) == 0) {
	Imported i = (Imported) iluparser_Malloc(sizeof *i);

	i->name = name;
	i->filename = 0;
	list_insert(imports, i);
    }
}


typedef struct {
    list imports;
    list sortedTypes;
} PtRock;


static void
processType(Type t, PtRock * r)
{
    static list pendingTypes;

    if (t->importInterfaceName != 0) {
	Type urt = myUrType(t);
	if (urt->interface) {
            addInterfaceToImports(r->imports, urt->interface);
	}
	return;
    }

    if (list_find(r->sortedTypes, matchPointer, t) != 0)
	return;
    if (pendingTypes == 0)
	pendingTypes = new_list();
    if (list_find(pendingTypes, matchPointer, t) != 0)
	fatalError("internal error: circularity in types");
    list_insert(pendingTypes, t);

    if (t->description) {
	switch (t->description->type) {
            case object_Type:
		{
                    Class c = t->description->structuredDes.object;

                    list_enumerate(c->superclasses, (EnumProc) processType,
				   r);
		}
		break;

            default:
		/* null */
		break;
	}
    }
    else {
	/* alias */
	processType(t->supertype, r);
    }

    list_remove(pendingTypes, t);
    list_insert(r->sortedTypes, t);
}


static void
normalizeInterface(Interface ifc)
{
    PtRock rock;
    rock.imports = ifc->imports;
    rock.sortedTypes = new_list();
    list_enumerate(ifc->types, (EnumProc) processType, &rock);
    list_clear(ifc->types, FALSE);
    ifc->types = rock.sortedTypes;
}


static void
generate(Interface ifc)
{
    IHandle ih = getContext(ifc);
    if (ih->forbidden) {
        fatalError(cat2("Must not generate stubs for ", ih->islIfName));
    }
    currentIfc = ifc;
    normalizeInterface(ifc);
    generateStub(ifc);
    generateLoadClass(ifc);
}


typedef enum {
    OptSpecial,
    OptDebug, 
    OptNoReport, 
    OptReportFile, 
    OptDir, OptDir1, 
    OptExSuffix, OptExSuffix1, 
    OptXrx, OptXrx1, 
    OptOmg, OptOmg1, 
    OptFlat, OptFlat1, 
    OptPrefix, OptPrefix1, 
    OptNoPrefix, OptNoPrefix1,
    OptHlp, OptHlp1, 
    OptNoHlp, OptNoHlp1,
    OptNoServ, OptNoServ1, 
    OptDel, OptDel1, 
    OptNull
} Option;


typedef struct {
    char *optName;
    Option opt;
    boolean ifaceArg;
    int otherArgs;
} OptTuple;


static OptTuple optionTable[] =
{
    {"null", OptNull, 0, 0}, /* must be first */
    {"special", OptSpecial, 0, 0},
    {"debug", OptDebug, 0, 0},
    {"noreport", OptNoReport, 0, 0},
    {"reportfile", OptReportFile, 0, 0},
    {"dir", OptDir, 0, 1},
    {"dir1", OptDir1, 1, 1},
    {"exsuffix", OptExSuffix, 0, 1},
    {"exsuffix1", OptExSuffix1, 1, 1},
    {"xrx", OptXrx, 0, 0},
    {"xrx1", OptXrx1, 1, 0},
    {"omg", OptOmg, 0, 0},
    {"omg1", OptOmg1, 1, 0},
    {"flat", OptFlat, 0, 0},
    {"flat", OptFlat1, 1, 0},
    {"prefix", OptPrefix, 0, 1},
    {"prefix1", OptPrefix1, 1, 1},
    {"noprefix", OptNoPrefix, 0, 0},
    {"noprefix1", OptNoPrefix1, 1, 0},
    {"hlp", OptHlp, 0, 0},
    {"hlp1", OptHlp1, 1, 0},
    {"nohlp", OptNoHlp, 0, 0},
    {"nohlp1", OptNoHlp1, 1, 0},
    {"noserv", OptNoServ, 0, 0},
    {"noserv1", OptNoServ1, 1, 0},
    {"delegation", OptDel, 0, 0},
    {"delegation1", OptDel1, 1, 0}
};


static void
printOptions()
{
    int i;
    for (i = 1; i < sizeof optionTable / sizeof optionTable[0]; i++) {
	fprintf(stderr, "[-%s", optionTable[i].optName);
	if (optionTable[i].ifaceArg) {
	    fprintf(stderr, " iface");
	}
	if (optionTable[i].otherArgs) {
	    fprintf(stderr, " arg");
	}
	fprintf(stderr, "] ");
    }
}


static int
getOptionTabIndex(char *name)
{
    int i;
    for (i = 1; i < sizeof optionTable / sizeof optionTable[0]; i++) {
	if (strcmp(name, optionTable[i].optName) == 0) return i;
    }
    return 0;
}

static Option
optionFromIndex(int i)
{
    return optionTable[i].opt;
}


static boolean
needsInterfaceFromIndex(int i)
{
    return optionTable[i].ifaceArg;
}


static void
usage(void)
{
    fprintf(stderr, "usage: %s ", programName);
    printOptions();
    fprintf(stderr, "files\n");
    exit(0);
}


static IHandle defaultIHandle = 0;


static IHandle allocIHandle(char* islifname) 
{
    IHandle ih = newContext();
    if (defaultIHandle) {
       list x = ih->stuffToLoad;
       *ih = *defaultIHandle;
       ih->stuffToLoad = x;
    }
    ih->islIfName = islifname;
    return ih;
}


static void contextAdd_xrx(IHandle ih)
{
    ih->stubSuffix = "_stub";
    ih->refSuffix = "";
    ih->visiHlp = 1;
    ih->sunHld = 0;
    ih->holderSuffix = "_var";
    ih->helperSuffix = "_help";
    ih->delClassSuffix = "_delegation";
}


static void contextAdd_Omg(IHandle ih)
{
    ih->stubSuffix = "Stub";
    ih->refSuffix = "";
    ih->visiHlp = 1;
    ih->sunHld = 1;
    ih->holderSuffix = "Holder";
    ih->helperSuffix = "Helper";
    ih->delClassSuffix = "_delegation";
}


static void contextAdd_flatness(IHandle ih, boolean flatness)
{
    ih->flatDir = flatness;
}


static void contextAdd_prefix(IHandle ih, char * prefix)
{
    if (stringlength(prefix)==0) prefix = 0;
    ih->prefix = prefix;
}


static void contextAdd_dirName(IHandle ih, char * dirName)
{
    if (stringlength(dirName)==0) dirName = 0;
    ih->dirName = dirName;
}


static void contextAdd_ExSuffix(IHandle ih, char * suffix)
{
    if (stringlength(suffix)==0) suffix = 0;
    ih->exceptSuffix = suffix;
}


int main(int argc, char **argv)
{
    boolean reportFlag = TRUE;
    char * reportFilename = 0;
    int optind = 1;
    char * ifname;
    IHandle ctx;
    initName();
    defaultIHandle = allocIHandle(0);
      contextAdd_xrx(defaultIHandle);
      defaultIHandle->prefix = "stubs";
      defaultIHandle->dirName = "javastubs";
      defaultIHandle->loaderClass_Name = "_allJavaStubs";
      setDefaultContext(defaultIHandle);
    {
        IHandle ctx = getContextN("ilu");
        ctx->forbidden = 1;
        contextAdd_prefix(ctx, "xerox");
    }
    programName = iluparser_GetProgramName(argv[0]);
    while (optind < argc && argv[optind][0] == '-') {
	char *opt = (argv[optind++])+1;
	int tabIndex = getOptionTabIndex(opt);
	ifname = 0;
	ctx = defaultIHandle;
	if (needsInterfaceFromIndex(tabIndex)) {
            if (optind>=argc) usage();
            ifname = argv[optind++]; 
            ctx = getContextN(ifname); 
	}
	switch (optionFromIndex(tabIndex)) {
            case OptDebug:
                debugFlag = TRUE;
                break;
            case OptNoReport:
                reportFlag = FALSE;
                break;
            case OptReportFile:
                reportFlag = TRUE;
                if (optind >= argc) usage();
                reportFilename = argv[optind++];
                reportFlag = (stringlength(reportFilename) > 0);
                break;
            case OptSpecial:
                {   
                    IHandle ctx = getContextN("ilu");
                    ctx->forbidden = 0;
                    ctx->visiHlp = FALSE;
                    ctx->loaderClass_Name = 0;
                }
                break;
            case OptFlat:
            case OptFlat1:
		contextAdd_flatness(ctx, 1);
		break;
            case OptNoPrefix:
            case OptNoPrefix1:
		contextAdd_prefix(ctx, 0);
		break;
            case OptPrefix:
            case OptPrefix1:
		if (optind>=argc) usage();
		contextAdd_prefix(ctx, argv[optind++]);
		break;
            case OptHlp:
            case OptHlp1:
		ctx->visiHlp = TRUE;
		break;
            case OptNoHlp:
            case OptNoHlp1:
		ctx->visiHlp = FALSE;
		break;
            case OptDir:
            case OptDir1:
		if (optind>=argc) usage();
		contextAdd_dirName(ctx, argv[optind++]);
		break;
            case OptExSuffix:
            case OptExSuffix1:
		if (optind>=argc) usage();
		contextAdd_ExSuffix(ctx, argv[optind++]);
		break;
            case OptOmg:
            case OptOmg1:
		contextAdd_Omg(ctx);
		break;
            case OptXrx:
            case OptXrx1:
		contextAdd_xrx(ctx);
		break;
            case OptNoServ:
            case OptNoServ1:
		ctx->noServer = 1;
		break;
            case OptDel:
            case OptDel1:
		ctx->useDel = 1;
		break;
            default:
		usage();
	}
    }
    setDefaultContext(defaultIHandle);
    while (optind < argc) {
	char *fileName = argv[optind++];
	list s;
	if ((s = ParseFile(fileName)) == 0) {
            fatalError("Couldn't find or parse %s.\n", fileName);
	}
	fprintf(stderr, "done with parsing\n");
	list_enumerate(s, (EnumProc) generate, NULL);
    }
    if (reportFlag) {
        ReportFiles(reportFilename);
    }
    return 0;
}


/* end */
